# pylint: disable=too-few-public-methods

"""A library which provides the HealthChecker class."""
from configparser import NoOptionError
import os
import sys
import logging
from queue import Queue

from anycast_healthchecker import PROGRAM_NAME
from anycast_healthchecker.servicecheck import ServiceCheck
from anycast_healthchecker.utils import (SERVICE_OPTIONS_TYPE,
                                         get_ip_prefixes_from_bird,
                                         get_ip_prefixes_from_config,
                                         reconfigure_bird,
                                         write_temp_bird_conf,
                                         archive_bird_conf,
                                         ServiceCheckDiedError,
                                         run_custom_bird_reconfigure)


class HealthChecker:
    """Launch threads for each service check and reconfigure BIRD daemon.

    It starts a thread for each service check we have in the configuration and
    then waits for reconfiguring Bird daemon based on the results of the
    service checks.

    It uses a queue as a way to communicate with all threads. Each thread will
    add an item in the queue, which contains the IP prefix to remove from or to
    add to BIRD configuration file. When item is added we pick it up, adjust
    BIRD configuration and then reload BIRD.

    This class should be instantiated once.

    Arguments:
        config (configparger obj): A configparser object with the configuration
        bird_configuration (dict): A dictionary with Bird settings.

    Methods:
        run(): Launches checks and updates BIRD configuration based on
        the result of the check.

    """

    def __init__(self, config, bird_configuration):
        """Initialization."""
        self.log = logging.getLogger(PROGRAM_NAME)
        self.config = config
        # A queue with IP prefixes and their action to be taken based on the
        # state of health check. An item is a tuple of 3 elements:
        #     1st: name of the thread.
        #     2nd: IP prefix.
        #     3nd: IP version, either '4' or '6'.
        self.action = Queue()
        self.bird_configuration = bird_configuration
        self.log.debug(self.bird_configuration)

        # A list of service checks
        self.services = config.sections()
        self.services.remove('daemon')

        # Holds IP prefixes per IP version for which we have a service check
        self.ip_prefixes = {}
        for ip_version in self.bird_configuration:
            _ip_prefixes = get_ip_prefixes_from_config(
                self.config,
                self.services,
                ip_version)

            _ip_prefixes.add(
                self.bird_configuration[ip_version]['dummy_ip_prefix']
            )
            self.ip_prefixes[ip_version] = _ip_prefixes

        self.log.info('initialize healthchecker')

    def _update_bird_conf_file(self, operation):
        """Update BIRD configuration.

        It adds to or removes IP prefix from BIRD configuration. It also
        updates generation time stamp in the configuration file.

        Main program will exit if configuration file cant be read/written.

        Arguments:
            operation (obj): Either an AddOperation or DeleteOperation object

        Returns:
            True if BIRD configuration was updated otherwise False.

        """
        conf_updated = False
        prefixes = []
        ip_version = operation.ip_version
        config_file = self.bird_configuration[ip_version]['config_file']
        variable_name = self.bird_configuration[ip_version]['variable_name']
        changes_counter =\
            self.bird_configuration[ip_version]['changes_counter']
        dummy_ip_prefix =\
            self.bird_configuration[ip_version]['dummy_ip_prefix']

        try:
            prefixes = get_ip_prefixes_from_bird(config_file)
        except OSError as error:
            self.log.error("failed to open Bird configuration %s, this is a "
                           "FATAL error, thus exiting main program", error)
            sys.exit(1)

        if not prefixes:
            self.log.error("found empty bird configuration %s, this is a FATAL"
                           " error, thus exiting main program", config_file)
            sys.exit(1)

        if dummy_ip_prefix not in prefixes:
            self.log.warning("dummy IP prefix %s wasn't found in bird "
                             "configuration, adding it. This shouldn't have "
                             "happened!", dummy_ip_prefix)
            prefixes.insert(0, dummy_ip_prefix)
            conf_updated = True

        ip_prefixes_without_check = set(prefixes).difference(
            self.ip_prefixes[ip_version])
        if ip_prefixes_without_check:
            self.log.warning("found %s IP prefixes in Bird configuration but "
                             "we aren't configured to run health checks on "
                             "them. Either someone modified the configuration "
                             "manually or something went horrible wrong. We "
                             "remove them from Bird configuration",
                             ','.join(ip_prefixes_without_check))
            # This is faster than using lambda and filter.
            # NOTE: We don't use remove method as we want to remove more than
            # occurrences of the IP prefixes without check.
            prefixes[:] = (ip for ip in prefixes
                           if ip not in ip_prefixes_without_check)
            conf_updated = True

        # Update the list of IP prefixes based on the status of health check.
        if operation.update(prefixes):
            conf_updated = True

        if not conf_updated:
            self.log.info('no updates for bird configuration')
            return conf_updated

        if self.bird_configuration[ip_version]['keep_changes']:
            archive_bird_conf(config_file, changes_counter)

        # some IP prefixes are either removed or added, create
        # configuration with new data.
        tempname = write_temp_bird_conf(
            dummy_ip_prefix,
            config_file,
            variable_name,
            prefixes
        )
        try:
            os.rename(tempname, config_file)
        except OSError as error:
            self.log.critical("failed to create Bird configuration %s, this "
                              "is a FATAL error, thus exiting main program",
                              error)
            sys.exit(1)
        else:
            self.log.info("Bird configuration for IPv%s is updated",
                          ip_version)

        # dummy_ip_prefix is always there
        if len(prefixes) == 1:
            self.log.warning("Bird configuration doesn't have IP prefixes for "
                             "any of the services we monitor! It means local "
                             "node doesn't receive any traffic")

        return conf_updated

    def run(self):
        """Launch checks and triggers updates on BIRD configuration."""
        # Launch a thread for each configuration
        if not self.services:
            self.log.warning("no service checks are configured")
        else:
            self.log.info("going to launch %s threads", len(self.services))
            if self.config.has_option('daemon', 'splay_startup'):
                splay_startup = self.config.getfloat('daemon', 'splay_startup')
            else:
                splay_startup = None

            for service in self.services:
                self.log.debug("launching thread for %s", service)
                _config = {}
                for option, getter in SERVICE_OPTIONS_TYPE.items():
                    try:
                        _config[option] = getattr(self.config, getter)(service,
                                                                       option)
                    except NoOptionError:
                        pass  # for optional settings

                _thread = ServiceCheck(service, _config, self.action,
                                       splay_startup)
                _thread.start()

        # Stay running until we are stopped
        while True:
            # Fetch items from action queue
            operation = self.action.get(block=True)

            if isinstance(operation, ServiceCheckDiedError):
                self.log.critical(operation)
                self.log.critical("This is a fatal error and the only way to "
                                  "recover is to restart, thus exiting with a "
                                  "non-zero code and let systemd act by "
                                  "triggering a restart")
                sys.exit(1)

            self.log.info("returned an item from the queue for %s with IP "
                          "prefix %s and action to %s Bird configuration",
                          operation.name,
                          operation.ip_prefix,
                          operation)
            bird_updated = self._update_bird_conf_file(operation)
            self.action.task_done()
            if bird_updated:
                ip_version = operation.ip_version
                if operation.bird_reconfigure_cmd is None:
                    reconfigure_bird(
                        self.bird_configuration[ip_version]['reconfigure_cmd'])
                else:
                    run_custom_bird_reconfigure(operation)



# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-instance-attributes
#

"""A library which provides the ServiceCheck class."""

import subprocess
import time
import logging
import traceback
from threading import Thread
import ipaddress
import random
import shlex

from anycast_healthchecker import PROGRAM_NAME
from anycast_healthchecker.utils import (AddOperation, DeleteOperation,
                                         ServiceCheckDiedError)


class ServiceCheck(Thread):
    """Handle the health checking for a service.

    Arguments:
        service (str): The name of the service to monitor.
        config (dict): A dictionary with the configuration of the service.
        action (Queue obj): A queue object to place actions based on the result
        of the health check.
        splay_startup: (float): The maximum time to delay the startup.

    Methods:
        run(): Run method of the thread.

    """

    def __init__(self, service, config, action, splay_startup):
        """Set the name of thread to be the name of the service."""
        super(ServiceCheck, self).__init__()
        self.name = service  # Used by Thread()
        self.daemon = True   # Used by Thread()
        self.config = config
        self.action = action
        self.splay_startup = splay_startup
        # sanity check has already been done, so the following *should* not
        # raise an exception
        _ip_prefix = ipaddress.ip_network(self.config['ip_prefix'])
        # NOTE: When subnetmask isn't provided ipaddress module creates an
        # object with a mask of /32 for IPv4 addresses and mask of /128 for
        # IPv6 addresses. As a result the prefix length is either 32 or 128
        # and we can get the IP address by looking at the network_address
        # attribute.
        self.ip_address = str(_ip_prefix.network_address)
        self.prefix_length = _ip_prefix.prefixlen
        self.ip_with_prefixlen = _ip_prefix.with_prefixlen
        self.ip_version = _ip_prefix.version
        self.ip_check_disabled = self.config['ip_check_disabled']
        self.log = logging.getLogger(PROGRAM_NAME)
        self.extra = {
            'ip_address': self.ip_address,
            'prefix_length': self.prefix_length,
            'ip_check_disabled': self.ip_check_disabled,
            'status': 'unknown',
        }
        self.add_operation = AddOperation(
            name=self.name,
            ip_prefix=self.ip_with_prefixlen,
            ip_version=self.ip_version,
            bird_reconfigure_timeout=(
                config['custom_bird_reconfigure_cmd_timeout']
            ),
            bird_reconfigure_cmd=config.get('custom_bird_reconfigure_cmd',
                                            None)
        )
        self.del_operation = DeleteOperation(
            name=self.name,
            ip_prefix=self.ip_with_prefixlen,
            ip_version=self.ip_version,
            bird_reconfigure_timeout=(
                config['custom_bird_reconfigure_cmd_timeout']
            ),
            bird_reconfigure_cmd=config.get('custom_bird_reconfigure_cmd',
                                            None)
        )
        self.log.info("loading check for %s", self.name, extra=self.extra)

    def _run_check(self):
        """Execute a check command.

        Returns:
            True if the exit code of the command is 0 otherwise False.

        """
        cmd = shlex.split(self.config['check_cmd'])
        self.log.info("running %s", ' '.join(cmd))
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)

        start_time = time.time()
        try:
            outs, errs = proc.communicate(timeout=self.config['check_timeout'])
        except subprocess.TimeoutExpired:
            self.log.error("check timed out")
            if proc.poll() is None:
                try:
                    proc.kill()
                except PermissionError:
                    self.log.warning("failed to kill check due to adequate "
                                     "access rights, check could be running "
                                     "under another user(root) via sudo")

            return False
        else:
            msg = "check duration {t:.3f}ms".format(
                t=(time.time() - start_time) * 1000)
            self.log.info(msg)

            if proc.returncode != 0:
                self.log.info("stderr from the check %s", errs)
                self.log.info("stdout from the check %s", outs)

            return proc.returncode == 0

    def _ip_assigned(self):
        """Check if IP prefix is assigned to loopback interface.

        Returns:
            True if IP prefix found assigned otherwise False.

        """
        output = []
        cmd = [
            '/sbin/ip',
            'address',
            'show',
            'dev',
            self.config['interface'],
            'to',
            self.ip_with_prefixlen,
        ]

        if self.ip_check_disabled:
            self.log.info("checking for IP assignment on interface %s is "
                          "disabled", self.config['interface'])
            return True

        self.log.debug("running %s", ' '.join(cmd))
        try:
            output = subprocess.check_output(
                cmd,
                universal_newlines=True,
                timeout=1)
        except subprocess.CalledProcessError as error:
            self.log.error("error checking IP-PREFIX %s: %s",
                           cmd, error.output)
            # Because it is unlikely to ever get an error we return True
            return True
        except subprocess.TimeoutExpired:
            self.log.error("timeout running %s", ' '.join(cmd))
            # Because it is unlikely to ever get a timeout we return True
            return True
        except ValueError as error:
            # We have been getting intermittent ValueErrors, see here
            # gist.github.com/unixsurfer/67db620d87f667423f6f6e3a04e0bff5
            # It has happened ~5 times and this code is executed from multiple
            # threads and every ~10secs on several (~40) production servers for
            # more than 18months.
            # It could be a bug in Python or system returns corrupted data.
            # As a consequence of the raised exception thread dies and the
            # service isn't monitored anymore!. So, we now catch the exception.
            # While checking if an IP is assigned, we get an error unrelated to
            # that prevents us from knowing if it's assigned. We simply don't
            # know. A retry logic could be a more proper solution.
            self.log.error("running %s raised ValueError exception:%s",
                           ' '.join(cmd), error)
            return True
        else:
            if self.ip_with_prefixlen in output:  # pylint: disable=E1135,R1705
                msg = "{i} assigned to loopback interface".format(
                    i=self.ip_with_prefixlen)
                self.log.debug(msg)
                return True
            else:
                msg = ("{i} isn't assigned to {d} interface"
                       .format(i=self.ip_with_prefixlen,
                               d=self.config['interface']))
                self.log.warning(msg)
                return False

        self.log.debug("I shouldn't land here!, it is a BUG")

        return False

    def _check_disabled(self):
        """Check if health check is disabled.

        It logs a message if health check is disabled and it also adds an item
        to the action queue based on 'on_disabled' setting.

        Returns:
            True if check is disabled otherwise False.

        """
        if self.config['check_disabled']:
            if self.config['on_disabled'] == 'withdraw':
                self.log.info("Check is disabled and ip_prefix will be "
                              "withdrawn")
                self.log.info("adding %s in the queue", self.ip_with_prefixlen)
                self.action.put(self.del_operation)
                self.log.info("Check is now permanently disabled")
            elif self.config['on_disabled'] == 'advertise':
                self.log.info("check is disabled, ip_prefix wont be withdrawn")
                self.log.info("adding %s in the queue", self.ip_with_prefixlen)
                self.action.put(self.add_operation)
                self.log.info('check is now permanently disabled')

            return True

        return False

    def run(self):
        """Wrap _run method."""
        # Catch all possible exceptions raised by the running thread
        # and let parent process know about it.
        try:
            self._run()
        except Exception:  # pylint: disable=broad-except
            self.action.put(
                ServiceCheckDiedError(self.name, traceback.format_exc())
            )

    def _run(self):
        """Discovers the health of a service.

        Runs until it is being killed from main program and is responsible to
        put an item into the queue based on the status of the health check.
        The status of service is consider UP after a number of consecutive
        successful health checks, in that case it asks main program to add the
        IP prefix associated with service to BIRD configuration, otherwise ask
        for a removal.
        Rise and fail options prevent unnecessary configuration changes when
        check is flapping.
        """
        up_cnt = 0
        down_cnt = 0
        # The current established state of the service check, it can be
        # either UP or DOWN but only after a number of consecutive successful
        # or unsuccessful health checks.
        check_state = 'Unknown'

        for key, value in self.config.items():
            self.log.debug("%s=%s:%s", key, value, type(value))

        # Service check will abort if it is disabled.
        if self._check_disabled():
            return

        if self.splay_startup is not None:
            sleep_time = float("%.3f" % random.uniform(0, self.splay_startup))
            self.log.info("delaying startup for %ssecs", sleep_time)
            time.sleep(sleep_time)

        interval = self.config['check_interval']
        start_offset = time.time() % interval
        # Go in a loop until we are told to stop
        while True:
            timestamp = time.time()
            if not self._ip_assigned():
                up_cnt = 0
                self.extra['status'] = 'down'
                self.log.warning("status DOWN because %s isn't assigned to "
                                 "loopback interface.",
                                 self.ip_with_prefixlen,
                                 extra=self.extra)
                if check_state != 'DOWN':
                    check_state = 'DOWN'
                    self.log.info("adding %s in the queue",
                                  self.ip_with_prefixlen,
                                  extra=self.extra)
                    self.action.put(self.del_operation)
            elif self._run_check():
                if up_cnt == (self.config['check_rise'] - 1):
                    self.extra['status'] = 'up'
                    self.log.info("status UP", extra=self.extra)
                    # Service exceeded all consecutive checks. Set its state
                    # accordingly and put an item in queue. But do it only if
                    # previous state was different, to prevent unnecessary bird
                    # reloads when a service flaps between states.
                    if check_state != 'UP':
                        check_state = 'UP'
                        self.log.info("adding %s in the queue",
                                      self.ip_with_prefixlen,
                                      extra=self.extra)
                        self.action.put(self.add_operation)
                elif up_cnt < self.config['check_rise']:
                    up_cnt += 1
                    self.log.info("going up %s", up_cnt, extra=self.extra)
                else:
                    self.log.error("up_cnt is higher %s, it's a BUG!",
                                   up_cnt,
                                   extra=self.extra)
                down_cnt = 0
            else:
                if down_cnt == (self.config['check_fail'] - 1):
                    self.extra['status'] = 'down'
                    self.log.info("status DOWN", extra=self.extra)
                    # Service exceeded all consecutive checks.
                    # Set its state accordingly and put an item in queue.
                    # But do it only if previous state was different, to
                    # prevent unnecessary bird reloads when a service flaps
                    # between states
                    if check_state != 'DOWN':
                        check_state = 'DOWN'
                        self.log.info("adding %s in the queue",
                                      self.ip_with_prefixlen,
                                      extra=self.extra)
                        self.action.put(self.del_operation)
                elif down_cnt < self.config['check_fail']:
                    down_cnt += 1
                    self.log.info("going down %s", down_cnt, extra=self.extra)
                else:
                    self.log.error("up_cnt is higher %s, it's a BUG!",
                                   up_cnt,
                                   extra=self.extra)
                up_cnt = 0

            self.log.info("wall clock time %.3fms",
                          (time.time() - timestamp) * 1000,
                          extra=self.extra)

            # calculate sleep time
            sleep = start_offset - time.time() % interval
            if sleep < 0:
                sleep += interval
            self.log.debug("sleeping for %.3fsecs", sleep, extra=self.extra)
            time.sleep(sleep)



# vim:fenc=utf-8
#
"""A healthchecker for Anycasted services."""
__title__ = 'anycast_healthchecker'
__author__ = 'Pavlos Parissis'
__license__ = 'Apache 2.0'
__version__ = '0.9.1'
__copyright__ = 'Copyright 2015-2020 Pavlos Parissis'

PROGRAM_NAME = __title__.replace('_', '-')

DEFAULT_OPTIONS = {
    'DEFAULT': {
        'interface': 'lo',
        'check_interval': 10,
        'check_timeout': 2,
        'check_rise': 2,
        'check_fail': 2,
        'check_disabled': 'true',
        'on_disabled': 'withdraw',
        'ip_check_disabled': 'false',
        'custom_bird_reconfigure_cmd_timeout': 2,
    },
    'daemon': {
        'ipv4': 'true',
        'ipv6': 'false',
        'bird_conf': '/var/lib/anycast-healthchecker/anycast-prefixes.conf',
        'bird6_conf': '/var/lib/anycast-healthchecker/6/anycast-prefixes.conf',
        'bird_variable': 'ACAST_PS_ADVERTISE',
        'bird6_variable': 'ACAST6_PS_ADVERTISE',
        'bird_reconfigure_cmd': 'sudo /usr/sbin/birdc configure',
        'bird6_reconfigure_cmd': 'sudo /usr/sbin/birdc6 configure',
        'dummy_ip_prefix': '10.189.200.255/32',
        'dummy_ip6_prefix': '2001:db8::1/128',
        'bird_keep_changes': 'false',
        'bird6_keep_changes': 'false',
        'bird_changes_counter': 128,
        'bird6_changes_counter': 128,
        'purge_ip_prefixes': 'false',
        'pidfile': '/var/run/anycast-healthchecker/anycast-healthchecker.pid',
        'loglevel': 'debug',
        'log_server_port': 514,
        'json_stdout': 'false',
        'json_log_file': 'false',
        'json_log_server': 'false',
        'log_maxbytes': 104857600,
        'log_backups': 8,
    }
}



# vim:fenc=utf-8
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-lines
"""Provide functions and classes that are used within anycast_healthchecker."""
from collections import Counter
import re
import os
import signal
import sys
import subprocess
import logging
import logging.handlers
import time
import datetime
import configparser
import glob
import shlex
import shutil
import ipaddress

from pythonjsonlogger import jsonlogger

from anycast_healthchecker import DEFAULT_OPTIONS, PROGRAM_NAME, __version__

SERVICE_OPTIONS_TYPE = {
    'check_cmd': 'get',
    'check_interval': 'getfloat',
    'check_timeout': 'getfloat',
    'check_rise': 'getint',
    'check_fail': 'getint',
    'check_disabled': 'getboolean',
    'on_disabled': 'get',
    'ip_prefix': 'get',
    'interface': 'get',
    'ip_check_disabled': 'getboolean',
    'custom_bird_reconfigure_cmd_timeout': 'getfloat',
    'custom_bird_reconfigure_cmd': 'get',
}
SERVICE_OPTIONAL_OPTIONS = {
    'custom_bird_reconfigure_cmd_timeout',
    'custom_bird_reconfigure_cmd',
}

DAEMON_OPTIONS_TYPE = {
    'pidfile': 'get',
    'bird_conf': 'get',
    'bird6_conf': 'get',
    'bird_variable': 'get',
    'bird6_variable': 'get',
    'log_maxbytes': 'getint',
    'log_backups': 'getint',
    'log_file': 'get',
    'stderr_file': 'get',
    'stderr_log_server': 'getboolean',
    'log_server': 'get',
    'log_server_port': 'getint',
    'json_stdout': 'getboolean',
    'json_log_server': 'getboolean',
    'json_log_file': 'getboolean',
    'purge_ip_prefixes': 'getboolean',
    'bird_keep_changes': 'getboolean',
    'bird6_keep_changes': 'getboolean',
    'bird_changes_counter': 'getint',
    'bird6_changes_counter': 'getint',
    'bird_reconfigure_cmd': 'get',
    'bird6_reconfigure_cmd': 'get',
    'splay_startup': 'getfloat',
}
DAEMON_OPTIONAL_OPTIONS = [
    'stderr_log_server',
    'stderr_file',
    'log_server',
    'log_file',
    'splay_startup',
]


def valid_ip_prefix(ip_prefix):
    """Perform a sanity check on ip_prefix.

    Arguments:
        ip_prefix (str): The IP-Prefix to validate

    Returns:
        True if ip_prefix is a valid IPv4 address with prefix length 32 or a
        valid IPv6 address with prefix length 128, otherwise False

    """
    try:
        ip_prefix = ipaddress.ip_network(ip_prefix)
    except ValueError:
        return False
    else:
        if ip_prefix.version == 4 and ip_prefix.max_prefixlen != 32:
            return False
        if ip_prefix.version == 6 and ip_prefix.max_prefixlen != 128:
            return False
        return True


def touch(file_path):
    """Touch a file in the same way as touch tool does.

    NOTE:
        If file_path doesn't exist it will be created.

    Arguments:
        file_path (str): The absolute file path

    Returns:
        None

    Raises:
        OSError exception

    """
    with open(file_path, 'a'):
        os.utime(file_path, None)


def get_ip_prefixes_from_config(config, services, ip_version):
    """Build a set of IP prefixes found in service configuration files.

    Arguments:
        config (obg): A configparser object which holds our configuration.
        services (list): A list of section names which are the name of the
        service checks.
        ip_version (int): IP protocol version

    Returns:
        A set of IP prefixes.

    """
    ip_prefixes = set()

    for service in services:
        ip_prefix = ipaddress.ip_network(config.get(service, 'ip_prefix'))
        if ip_prefix.version == ip_version:
            ip_prefixes.add(ip_prefix.with_prefixlen)

    return ip_prefixes


def ip_prefixes_sanity_check(config, bird_configuration):
    """Sanity check on IP prefixes.

    Arguments:
        config (obg): A configparser object which holds our configuration.
        bird_configuration (dict): A dictionary, which holds Bird configuration
        per IP protocol version.

    """
    for ip_version in bird_configuration:
        modify_ip_prefixes(config,
                           bird_configuration[ip_version]['config_file'],
                           bird_configuration[ip_version]['variable_name'],
                           bird_configuration[ip_version]['dummy_ip_prefix'],
                           bird_configuration[ip_version]['reconfigure_cmd'],
                           bird_configuration[ip_version]['keep_changes'],
                           bird_configuration[ip_version]['changes_counter'],
                           ip_version)


def modify_ip_prefixes(
        config,
        config_file,
        variable_name,
        dummy_ip_prefix,
        reconfigure_cmd,
        keep_changes,
        changes_counter,
        ip_version):
    """Modify IP prefixes in Bird configuration.

    Depending on the configuration either removes or reports IP prefixes found
    in Bird configuration for which we don't have a service check associated
    with them. Moreover, it adds the dummy IP prefix if it isn't present and
    ensures that the correct variable name is set.

    Arguments:
        config (obg): A configparser object which holds our configuration.
        config_file (str): The file name of bird configuration
        variable_name (str): The name of the variable set in bird configuration
        dummy_ip_prefix (str): The dummy IP prefix, which must be always
        reconfigure_cmd (str): The command to run to trigger a reconfiguration
        on Bird daemon upon successful configuration update
        keep_changes (boolean): To enable keeping a history of changes applied
        to bird configuration
        changes_counter (int): The number of configuration changes to keep
        ip_version (int): IP protocol version of Bird configuration

    """
    log = logging.getLogger(PROGRAM_NAME)
    services = config.sections()
    services.remove('daemon')  # not needed during sanity check for IP-Prefixes
    update_bird_conf = False
    try:
        ip_prefixes_in_bird = get_ip_prefixes_from_bird(config_file)
    except OSError as error:
        log.error("failed to open Bird configuration %s, this is a FATAL "
                  "error, thus exiting main program", error)
        sys.exit(1)

    _name = get_variable_name_from_bird(config_file)
    if _name is None:
        log.warning("failed to find variable name in %s, going to add it",
                    config_file)
        update_bird_conf = True
    elif _name != variable_name:
        log.warning("found incorrect variable name in %s, going to add the "
                    "correct one %s", _name, variable_name)
        update_bird_conf = True

    if dummy_ip_prefix not in ip_prefixes_in_bird:
        log.warning("dummy IP prefix %s is missing from bird configuration "
                    "%s, adding it", dummy_ip_prefix, config_file)
        ip_prefixes_in_bird.insert(0, dummy_ip_prefix)
        update_bird_conf = True

    # Find IP prefixes in Bird configuration without a check.
    ip_prefixes_with_check = get_ip_prefixes_from_config(
        config,
        services,
        ip_version)
    # dummy_ip_prefix doesn't have a config by design
    ip_prefixes_with_check.add(dummy_ip_prefix)

    ip_prefixes_without_check = set(ip_prefixes_in_bird).difference(
        ip_prefixes_with_check)

    if ip_prefixes_without_check:
        if config.getboolean('daemon', 'purge_ip_prefixes'):
            log.warning("removing IP prefix(es) %s from %s because they don't "
                        "have a service check configured",
                        ','.join(ip_prefixes_without_check),
                        config_file)
            ip_prefixes_in_bird[:] = (ip for ip in ip_prefixes_in_bird
                                      if ip not in ip_prefixes_without_check)
            update_bird_conf = True
        else:
            log.warning("found IP prefixes %s in %s without a service "
                        "check configured",
                        ','.join(ip_prefixes_without_check),
                        config_file)

    if update_bird_conf:
        if keep_changes:
            archive_bird_conf(config_file, changes_counter)
        tempname = write_temp_bird_conf(
            dummy_ip_prefix,
            config_file,
            variable_name,
            ip_prefixes_in_bird
        )
        try:
            os.rename(tempname, config_file)
        except OSError as error:
            msg = ("CRITICAL: failed to create Bird configuration {e}, "
                   "this is FATAL error, thus exiting main program"
                   .format(e=error))
            sys.exit("{m}".format(m=msg))
        else:
            log.info("Bird configuration for IPv%s is updated", ip_version)
            reconfigure_bird(reconfigure_cmd)


def load_configuration(config_file, config_dir, service_file):
    """Build configuration objects.

    If all sanity checks against daemon and service check settings are passed
    then it builds a ConfigParser object which holds all our configuration
    and a dictionary data structure which holds Bird configuration per IP
    protocol version.

    Arguments:
        config_file (str): The file name which holds daemon settings
        config_dir (str): The directory name which has configuration files
        for each service check
        service_file (str): A file which contains configuration for a single
        service check

    Returns:
        A tuple with 1st element a ConfigParser object and 2nd element
        a dictionary.
    Raises:
        ValueError if a sanity check fails.

    """
    config_files = [config_file]
    config = configparser.ConfigParser()
    config.read_dict(DEFAULT_OPTIONS)

    if not os.path.isfile(config_file):
        raise ValueError("{f} configuration file either isn't readable or "
                         "doesn't exist".format(f=config_file))
    if service_file is not None:
        if not os.path.isfile(service_file):
            raise ValueError("{f} configuration file for a service check "
                             "doesn't exist".format(f=service_file))
        else:
            config_files.append(service_file)
    elif config_dir is not None:
        if not os.path.isdir(config_dir):
            raise ValueError("{d} directory with configuration files for "
                             "service checks doesn't exist"
                             .format(d=config_dir))
        else:
            config_files.extend(glob.glob(os.path.join(config_dir, '*.conf')))

    try:
        config.read(config_files)
    except configparser.Error as exc:
        raise ValueError(exc)

    configuration_check(config)
    bird_configuration = build_bird_configuration(config)
    create_bird_config_files(bird_configuration)

    return config, bird_configuration


def configuration_check(config):
    """Perform a sanity check on configuration.

    First it performs a sanity check against settings for daemon
    and then against settings for each service check.

    Arguments:
        config (obj): A configparser object which holds our configuration.

    Returns:
        None if all checks are successfully passed otherwise raises a
        ValueError exception.

    """
    log_level = config.get('daemon', 'loglevel')
    num_level = getattr(logging, log_level.upper(), None)
    pidfile = config.get('daemon', 'pidfile')

    # Catch the case where the directory, under which we store the pid file, is
    # missing.
    if not os.path.isdir(os.path.dirname(pidfile)):
        raise ValueError("{d} doesn't exit".format(d=os.path.dirname(pidfile)))

    if not isinstance(num_level, int):
        raise ValueError('Invalid log level: {}'.format(log_level))

    for _file in 'log_file', 'stderr_file':
        if config.has_option('daemon', _file):
            try:
                touch(config.get('daemon', _file))
            except OSError as exc:
                raise ValueError(exc)

    for option, getter in DAEMON_OPTIONS_TYPE.items():
        try:
            getattr(config, getter)('daemon', option)
        except configparser.NoOptionError as error:
            if option not in DAEMON_OPTIONAL_OPTIONS:
                raise ValueError(error)
        except configparser.Error as error:
            raise ValueError(error)
        except ValueError as exc:
            msg = ("invalid data for '{opt}' option in daemon section: {err}"
                   .format(opt=option, err=exc))
            raise ValueError(msg)

    service_configuration_check(config)


def service_configuration_check(config):
    """Perform a sanity check against options for each service check.

    Arguments:
        config (obj): A configparser object which holds our configuration.

    Returns:
        None if all sanity checks are successfully passed otherwise raises a
        ValueError exception.

    """
    ipv4_enabled = config.getboolean('daemon', 'ipv4')
    ipv6_enabled = config.getboolean('daemon', 'ipv6')
    services = config.sections()
    # we don't need it during sanity check for services check
    services.remove('daemon')
    ip_prefixes = []

    for service in services:
        for option, getter in SERVICE_OPTIONS_TYPE.items():
            try:
                getattr(config, getter)(service, option)
            except configparser.NoOptionError as error:
                if option not in SERVICE_OPTIONAL_OPTIONS:
                    raise ValueError(error)
            except configparser.Error as error:
                raise ValueError(error)
            except ValueError as exc:
                msg = ("invalid data for '{opt}' option in service check "
                       "{name}: {err}"
                       .format(opt=option, name=service, err=exc))
                raise ValueError(msg)

        if (config.get(service, 'on_disabled') != 'withdraw' and
                config.get(service, 'on_disabled') != 'advertise'):
            msg = ("'on_disabled' option has invalid value ({val}) for "
                   "service check {name}, 'on_disabled option should be set "
                   "either to 'withdraw' or to 'advertise'"
                   .format(name=service,
                           val=config.get(service, 'on_disabled')))
            raise ValueError(msg)

        ip_prefixes.append(config.get(service, 'ip_prefix'))

        if not valid_ip_prefix(config.get(service, 'ip_prefix')):
            msg = ("invalid value ({val}) for 'ip_prefix' option in service "
                   "check {name}. It should be an IP PREFIX in form of "
                   "ip/prefixlen."
                   .format(name=service, val=config.get(service, 'ip_prefix')))
            raise ValueError(msg)

        _ip_prefix = ipaddress.ip_network(config.get(service, 'ip_prefix'))
        if not ipv6_enabled and _ip_prefix.version == 6:
            raise ValueError("IPv6 support is disabled in "
                             "anycast-healthchecker while there is an IPv6 "
                             "prefix configured for {name} service check"
                             .format(name=service))
        if not ipv4_enabled and _ip_prefix.version == 4:
            raise ValueError("IPv4 support is disabled in "
                             "anycast-healthchecker while there is an IPv4 "
                             "prefix configured for {name} service check"
                             .format(name=service))

        cmd = shlex.split(config.get(service, 'check_cmd'))
        try:
            proc = subprocess.Popen(cmd)
            proc.kill()
        except (OSError, subprocess.SubprocessError) as exc:
            msg = ("failed to run check command '{cmd}' for service check "
                   "{name}: {err}"
                   .format(name=service,
                           cmd=config.get(service, 'check_cmd'),
                           err=exc))
            raise ValueError(msg)

    occurrences_of_ip_prefixes = Counter(ip_prefixes)
    for ip_prefix, counter in occurrences_of_ip_prefixes.items():
        if counter > 1:
            raise ValueError("{ip} is used by {c} service checks"
                             .format(ip=ip_prefix, c=counter))


def build_bird_configuration(config):
    """Build bird configuration structure.

    First it performs a sanity check against bird settings and then builds a
    dictionary structure with bird configuration per IP version.

    Arguments:
        config (obj): A configparser object which holds our configuration.

    Returns:
        A dictionary

    Raises:
        ValueError if sanity check fails.

    """
    bird_configuration = {}

    if config.getboolean('daemon', 'ipv4'):
        if os.path.islink(config.get('daemon', 'bird_conf')):
            config_file = os.path.realpath(config.get('daemon', 'bird_conf'))
            print("'bird_conf' is set to a symbolic link ({s} -> {d}, but we "
                  "will use the canonical path of that link"
                  .format(s=config.get('daemon', 'bird_conf'), d=config_file))
        else:
            config_file = config.get('daemon', 'bird_conf')

        dummy_ip_prefix = config.get('daemon', 'dummy_ip_prefix')
        if not valid_ip_prefix(dummy_ip_prefix):
            raise ValueError("invalid dummy IPv4 prefix: {i}"
                             .format(i=dummy_ip_prefix))

        bird_configuration[4] = {
            'config_file': config_file,
            'variable_name': config.get('daemon', 'bird_variable'),
            'dummy_ip_prefix': dummy_ip_prefix,
            'reconfigure_cmd': config.get('daemon', 'bird_reconfigure_cmd'),
            'keep_changes': config.getboolean('daemon', 'bird_keep_changes'),
            'changes_counter': config.getint('daemon', 'bird_changes_counter')
        }
    if config.getboolean('daemon', 'ipv6'):
        if os.path.islink(config.get('daemon', 'bird6_conf')):
            config_file = os.path.realpath(config.get('daemon', 'bird6_conf'))
            print("'bird6_conf' is set to a symbolic link ({s} -> {d}, but we "
                  "will use the canonical path of that link"
                  .format(s=config.get('daemon', 'bird6_conf'), d=config_file))
        else:
            config_file = config.get('daemon', 'bird6_conf')

        dummy_ip_prefix = config.get('daemon', 'dummy_ip6_prefix')
        if not valid_ip_prefix(dummy_ip_prefix):
            raise ValueError("invalid dummy IPv6 prefix: {i}"
                             .format(i=dummy_ip_prefix))
        bird_configuration[6] = {
            'config_file': config_file,
            'variable_name': config.get('daemon', 'bird6_variable'),
            'dummy_ip_prefix': dummy_ip_prefix,
            'reconfigure_cmd': config.get('daemon', 'bird6_reconfigure_cmd'),
            'keep_changes': config.getboolean('daemon', 'bird6_keep_changes'),
            'changes_counter': config.getint('daemon', 'bird6_changes_counter')
        }

    return bird_configuration


def get_variable_name_from_bird(bird_conf):
    """Return the variable name set in Bird configuration.

    The variable name in Bird configuration is set with the keyword 'define',
    here is an example:

        define ACAST_PS_ADVERTISE =

    and we exract the string between the word 'define' and the equals sign.

    Arguments:
        bird_conf (str): The absolute file name path of Bird configuration.

    Returns:
        The variable name as a string or None if it isn't found.

    """
    bird_variable_pattern = re.compile(
        r'''
        ^\s*
        define\s+
        (?P<name>\S+\b)
        \s+
        =
        ''', re.VERBOSE
    )

    with open(bird_conf, 'r') as content:
        for line in content.readlines():
            variable_match = bird_variable_pattern.search(line)
            if variable_match:
                return variable_match.group('name')

    return None


def create_bird_config_files(bird_configuration):
    """Create bird configuration files per IP version.

    Creates bird configuration files if they don't exist. It also creates the
    directories where we store the history of changes, if this functionality is
    enabled.

    Arguments:
        bird_configuration (dict): A dictionary with settings for bird.

    Returns:
       None

    Raises:
        ValueError if we can't create bird configuration files and the
        directory to store the history of changes in bird configuration file.

    """
    for ip_version in bird_configuration:
        # This creates the file if it doesn't exist.
        config_file = bird_configuration[ip_version]['config_file']
        try:
            touch(config_file)
        except OSError as exc:
            raise ValueError("failed to create {f}:{e}"
                             .format(f=config_file, e=exc))
        if bird_configuration[ip_version]['keep_changes']:
            history_dir = os.path.join(os.path.dirname(config_file), 'history')
            try:
                os.mkdir(history_dir)
            except FileExistsError:
                pass
            except OSError as exc:
                raise ValueError("failed to make directory {d} for keeping a "
                                 "history of changes for {b}:{e}"
                                 .format(d=history_dir, b=config_file, e=exc))
            else:
                print("{d} is created".format(d=history_dir))


def running(processid):
    """Check the validity of a process ID.

    Arguments:
        processid (int): Process ID number.

    Returns:
        True if process ID is found otherwise False.

    """
    try:
        # From kill(2)
        #   If sig is 0 (the null signal), error checking is performed but no
        #   signal is actually sent. The null signal can be used to check the
        #   validity of pid
        os.kill(processid, 0)
    except OverflowError as exc:
        print("checking validity of pid ({p}) failed with: {e}"
              .format(p=processid, e=exc))
        sys.exit(1)
    except OSError:
        return False
    else:
        return True


def get_ip_prefixes_from_bird(filename):
    """Build a list of IP prefixes found in Bird configuration.

    Arguments:
        filename (str): The absolute path of the Bird configuration file.

    Notes:
        It can only parse a file with the following format

            define ACAST_PS_ADVERTISE =
                [
                    10.189.200.155/32,
                    10.189.200.255/32
                ];

    Returns:
        A list of IP prefixes.

    """
    prefixes = []
    with open(filename, 'r') as bird_conf:
        lines = bird_conf.read()

    for line in lines.splitlines():
        line = line.strip(', ')
        if valid_ip_prefix(line):
            prefixes.append(line)

    return prefixes


class BaseOperation(object):
    """Run operation on a list.

    Arguments:
        name (string): The name of the service for the given ip_prefix
        ip_prefix (string): The value to run the operation
        ip_version (string): IP protocol version
        bird_reconfigure_cmd (string): A custom command to trigger
        reconfiguration to Bird Daemon.
        bird_reconfigure_cmd_timeout (float): Maximum time to wait for command
        to complete.
    """

    def __init__(
            self,
            name,
            ip_prefix,
            ip_version,
            bird_reconfigure_cmd,
            bird_reconfigure_timeout):  # noqa:D102
        self.name = name
        self.ip_prefix = ip_prefix
        self.log = logging.getLogger(PROGRAM_NAME)
        self.ip_version = ip_version
        self.bird_reconfigure_cmd = bird_reconfigure_cmd
        self.bird_reconfigure_timeout = bird_reconfigure_timeout


class AddOperation(BaseOperation):
    """Add a value to a list."""

    def __str__(self):
        """Handy string representation."""
        return 'add to'

    def update(self, prefixes):
        """Add a value to the list.

        Arguments:
            prefixes(list): A list to add the value
        """
        if self.ip_prefix not in prefixes:
            prefixes.append(self.ip_prefix)
            self.log.info("announcing %s for %s", self.ip_prefix, self.name)
            return True

        return False


class DeleteOperation(BaseOperation):
    """Remove a value from a list."""

    def __str__(self):
        """Handy string representation."""
        return 'delete from'

    def update(self, prefixes):
        """Remove a value to the list.

        Arguments:
            prefixes(list): A list to remove the value
        """
        if self.ip_prefix in prefixes:
            prefixes.remove(self.ip_prefix)
            self.log.info("withdrawing %s for %s", self.ip_prefix, self.name)
            return True

        return False


def reconfigure_bird(cmd):
    """Reconfigure BIRD daemon.

    Arguments:
        cmd (string): A command to trigger a reconfiguration of Bird daemon

    Notes:
        Runs 'birdc configure' to reconfigure BIRD. Some useful information on
        how birdc tool works:
            -- Returns a non-zero exit code only when it can't access BIRD
               daemon via the control socket (/var/run/bird.ctl). This happens
               when BIRD daemon is either down or when the caller of birdc
               doesn't have access to the control socket.
            -- Returns zero exit code when reconfigure fails due to invalid
               configuration. Thus, we catch this case by looking at the output
               and not at the exit code.
            -- Returns zero exit code when reconfigure was successful.
            -- Should never timeout, if it does then it is a bug.

    """
    log = logging.getLogger(PROGRAM_NAME)
    cmd = shlex.split(cmd)
    log.info("reconfiguring BIRD by running %s", ' '.join(cmd))
    try:
        output = subprocess.check_output(
            cmd,
            timeout=2,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            )
    except subprocess.TimeoutExpired:
        log.error("reconfiguring bird timed out")
        return
    except subprocess.CalledProcessError as error:
        # birdc returns 0 even when it fails due to invalid config,
        # but it returns 1 when BIRD is down.
        log.error("reconfiguring BIRD failed, either BIRD daemon is down or "
                  "we don't have privileges to reconfigure it (sudo problems?)"
                  ":%s", error.output.strip())
        return
    except FileNotFoundError as error:
        log.error("reconfiguring BIRD failed with: %s", error)
        return

    # 'Reconfigured' string will be in the output if and only if conf is valid.
    pattern = re.compile('^Reconfigured$', re.MULTILINE)
    if pattern.search(str(output)):
        log.info('reconfigured BIRD daemon')
    else:
        # We will end up here only if we generated an invalid conf
        # or someone broke bird.conf.
        log.error("reconfiguring BIRD returned error, most likely we generated"
                  " an invalid configuration file or Bird configuration in is "
                  "broken:%s", output)


def write_temp_bird_conf(dummy_ip_prefix,
                         config_file,
                         variable_name,
                         prefixes):
    """Write in a temporary file the list of IP-Prefixes.

    A failure to create and write the temporary file will exit main program.

    Arguments:
        dummy_ip_prefix (str): The dummy IP prefix, which must be always
        config_file (str): The file name of bird configuration
        variable_name (str): The name of the variable set in bird configuration
        prefixes (list): The list of IP-Prefixes to write

    Returns:
        The filename of the temporary file

    """
    log = logging.getLogger(PROGRAM_NAME)
    comment = ("# {i} is a dummy IP Prefix. It should NOT be used and "
               "REMOVED from the constant.".format(i=dummy_ip_prefix))

    # the temporary file must be on the same filesystem as the bird config
    # as we use os.rename to perform an atomic update on the bird config.
    # Thus, we create it in the same directory that bird config is stored.
    tm_file = os.path.join(os.path.dirname(config_file), str(time.time()))
    log.debug("going to write to %s", tm_file)

    try:
        with open(tm_file, 'w') as tmpf:
            tmpf.write("# Generated {t} by {n} (pid={p})\n"
                       .format(t=datetime.datetime.now(),
                               n=PROGRAM_NAME,
                               p=os.getpid()))
            tmpf.write("{c}\n".format(c=comment))
            tmpf.write("define {n} =\n".format(n=variable_name))
            tmpf.write("{s}[\n".format(s=4 * ' '))
            # all entries of the array need a trailing comma except the last
            # one. A single element array doesn't need a trailing comma.
            tmpf.write(',\n'.join([' '*8 + n for n in prefixes]))
            tmpf.write("\n{s}];\n".format(s=4 * ' '))
    except OSError as error:
        log.critical("failed to write temporary file %s: %s. This is a FATAL "
                     "error, this exiting main program", tm_file, error)
        sys.exit(1)
    else:
        return tm_file


def archive_bird_conf(config_file, changes_counter):
    """Keep a history of Bird configuration files.

    Arguments:
        config_file (str): file name of bird configuration
        changes_counter (int): number of configuration files to keep in the
        history
    """
    log = logging.getLogger(PROGRAM_NAME)
    history_dir = os.path.join(os.path.dirname(config_file), 'history')
    dst = os.path.join(history_dir, str(time.time()))
    log.debug("coping %s to %s", config_file, dst)

    history = [x for x in os.listdir(history_dir)
               if os.path.isfile(os.path.join(history_dir, x))]

    if len(history) > changes_counter:
        log.info("threshold of %s is reached, removing old files",
                 changes_counter)
        for _file in sorted(history, reverse=True)[changes_counter - 1:]:
            _path = os.path.join(history_dir, _file)
            try:
                os.remove(_path)
            except OSError as exc:
                log.warning("failed to remove %s: %s", _file, exc)
            else:
                log.info("removed %s", _path)

    try:
        shutil.copy2(config_file, dst)
    except OSError as exc:
        log.warning("failed to copy %s to %s: %s", config_file, dst, exc)


def update_pidfile(pidfile):
    """Update pidfile.

    Notice:
        We should call this function only after we have successfully acquired
        a lock and never before. It exits main program if it fails to parse
        and/or write pidfile.

    Arguments:
        pidfile (str): pidfile to update

    """
    try:
        with open(pidfile, mode='r') as _file:
            pid = _file.read(1024).rstrip()

        try:
            pid = int(pid)
        except ValueError:
            print("cleaning stale pidfile with invalid data:'{}'".format(pid))
            write_pid(pidfile)
        else:
            if running(pid):
                # This is to catch migration issues from 0.7.x to 0.8.x
                # version, where old process is still around as it failed to
                # be stopped. Since newer version has a different locking
                # mechanism, we can end up with both versions running.
                # In order to avoid this situation we refuse to startup.
                sys.exit("process {} is already running".format(pid))
            else:
                # pidfile exists with a PID for a process that is not running.
                # Let's update PID.
                print("updating stale processID({}) in pidfile".format(pid))
                write_pid(pidfile)
    except FileNotFoundError:
        # Either it's 1st time we run or previous run was terminated
        # successfully.
        print("creating pidfile {f}".format(f=pidfile))
        write_pid(pidfile)
    except OSError as exc:
        sys.exit("failed to update pidfile:{e}".format(e=exc))


def write_pid(pidfile):
    """Write processID to the pidfile.

    Notice:
        It exits main program if it fails to write pidfile.

    Arguments:
        pidfile (str): pidfile to update

    """
    pid = str(os.getpid())
    try:
        with open(pidfile, mode='w') as _file:
            print("writing processID {p} to pidfile".format(p=pid))
            _file.write(pid)
    except OSError as exc:
        sys.exit("failed to write pidfile:{e}".format(e=exc))


def shutdown(pidfile, signalnb=None, frame=None):
    """Clean up pidfile upon shutdown.

    Notice:
        We should register this function as signal handler for the following
        termination signals:
            SIGHUP
            SIGTERM
            SIGABRT
            SIGINT

    Arguments:
        pidfile (str): pidfile to remove
        signalnb (int): The ID of signal
        frame (obj): Frame object at the time of receiving the signal

    """
    log = logging.getLogger(PROGRAM_NAME)
    log.info("received %s at %s", signalnb, frame)
    log.info("going to remove pidfile %s", pidfile)
    # no point to catch possible errors when we delete the pid file
    os.unlink(pidfile)
    log.info('shutdown is complete')
    sys.exit(0)


def setup_logger(config):
    """Configure the logging environment.

    Notice:
        By default logging will go to STDOUT and messages for unhandled
        exceptions or crashes will go to STDERR. If log_file and/or log_server
        is set then we don't log to STDOUT. Messages for unhandled exceptions
        or crashes can only go to either STDERR or to stderr_file or to
        stderr_log_server.

    Arguments:
        config (obj): A configparser object which holds our configuration.

    Returns:
        A logger with all possible handlers configured.

    """
    logger = logging.getLogger(PROGRAM_NAME)
    num_level = getattr(
        logging,
        config.get('daemon', 'loglevel').upper(),  # pylint: disable=no-member
        None
    )
    logger.setLevel(num_level)
    lengths = []
    for section in config:
        lengths.append(len(section))

    width = sorted(lengths)[-1] + 1

    def log_format():
        """Produce a log format line."""
        supported_keys = [
            'asctime',
            'levelname',
            'process',
            # 'funcName',
            # 'lineno',
            'threadName',
            'message',
        ]

        return ' '.join(['%({0:s})'.format(i) for i in supported_keys])

    custom_format = log_format()
    json_formatter = CustomJsonFormatter(custom_format,
                                         prefix=PROGRAM_NAME + ': ')
    formatter = logging.Formatter(
        '%(asctime)s {program}[%(process)d] %(levelname)-8s '
        '%(threadName)-{width}s %(message)s'
        .format(program=PROGRAM_NAME, width=width)
    )

    # Register logging handlers based on configuration.
    if config.has_option('daemon', 'log_file'):
        file_handler = logging.handlers.RotatingFileHandler(
            config.get('daemon', 'log_file'),
            maxBytes=config.getint('daemon', 'log_maxbytes'),
            backupCount=config.getint('daemon', 'log_backups')
        )

        if config.getboolean('daemon', 'json_log_file'):
            file_handler.setFormatter(json_formatter)
        else:
            file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    if config.has_option('daemon', 'log_server'):
        udp_handler = logging.handlers.SysLogHandler(
            (
                config.get('daemon', 'log_server'),
                config.getint('daemon', 'log_server_port')
            )
        )

        if config.getboolean('daemon', 'json_log_server'):
            udp_handler.setFormatter(json_formatter)
        else:
            udp_handler.setFormatter(formatter)
        logger.addHandler(udp_handler)

    # Log to STDOUT if and only if log_file and log_server aren't enabled
    if (not config.has_option('daemon', 'log_file')
            and not config.has_option('daemon', 'log_server')):
        stream_handler = logging.StreamHandler()
        if config.getboolean('daemon', 'json_stdout'):
            stream_handler.setFormatter(json_formatter)
        else:
            stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    # We can redirect STDERR only to one destination.
    if config.has_option('daemon', 'stderr_file'):
        sys.stderr = CustomRotatingFileLogger(
            filepath=config.get('daemon', 'stderr_file'),
            maxbytes=config.getint('daemon', 'log_maxbytes'),
            backupcount=config.getint('daemon', 'log_backups')
        )
    elif (config.has_option('daemon', 'stderr_log_server')
          and not config.has_option('daemon', 'stderr_file')):
        sys.stderr = CustomUdpLogger(
            server=config.get('daemon', 'log_server'),
            port=config.getint('daemon', 'log_server_port')
        )
    else:
        print('messages for unhandled exceptions will go to STDERR')

    return logger


class CustomLogger(object):
    """Helper Logger to redirect STDOUT or STDERR to a logging hander.

    It wraps a Logger class into a file like object, which provides a handy
    way to redirect STDOUT or STDERR to a logger. This class provides the
    necessary methods (write, flush and close) to build a file-like object
    and it can not be used directly as it does not provide a logging handler.
    Instead, you must instantiate one of subclass (CustomRotatingFileLogger and
    CustomUdpLogger).

    Arguments
        handler (int): A logging handler to use.

    Methods:
        write(string): Write string to logger with newlines removed.
        flush(): Flushe logger messages.
        close(): Close logger.

    Returns:
        A logger object.

    """

    def __init__(self, handler):
        """Create a logging.Logger class with extended functionality."""
        log_format = ('%(asctime)s {program}[%(process)d] '
                      '%(threadName)s %(message)s'
                      .format(program=PROGRAM_NAME))
        self.logger = logging.getLogger('stderr')
        self.logger.setLevel(logging.DEBUG)
        self.handler = handler
        formatter = logging.Formatter(log_format)
        self.handler.setFormatter(formatter)
        self.logger.addHandler(self.handler)

    def write(self, string):
        """Erase newline from a string and write to the logger."""
        string = string.rstrip()
        if string:  # Don't log empty lines
            self.logger.critical(string)

    def flush(self):
        """Flush logger's data."""
        # In case multiple handlers are attached to the logger make sure they
        # are flushed.
        for handler in self.logger.handlers:
            handler.flush()

    def close(self):
        """Call the closer method of the logger."""
        # In case multiple handlers are attached to the logger make sure they
        # are all closed.
        for handler in self.logger.handlers:
            handler.close()


class CustomRotatingFileLogger(CustomLogger):
    """Subclass CustomLogger to provide a rotating file logger.

    The rotation of log file is enabled by default.

    Usage:
        >>> import sys
        >>> sys.stderr = CustomRotatingFileLogger(filepath='/foo.log')

    Arguments:
        filepath (str) : file path of the log file
        maxbytes (int) : log file is rotated when it grows bigger than maxbytes
        backupcount (int): maximum rotated log file to keep.

    Notice:
        If maxbytes is zero, rollover never occurs and we may fill up the disk.
        An external program (logrotate) must not rotate the file in this case,
        as the file descriptor for the log file will be changed without us
        knowing this.

    Returns:
        A logger object.

    """

    def __init__(self, filepath, *, maxbytes=10485, backupcount=8):
        """Create a logging.Logger class with extended functionality."""
        handler = logging.handlers.RotatingFileHandler(filepath,
                                                       maxBytes=maxbytes,
                                                       backupCount=backupcount)
        super().__init__(handler=handler)


class CustomUdpLogger(CustomLogger):
    """Subclass CustomLogger to provide a UDP logger.

    Usage:
        >>> import sys
        >>> sys.stderr = CustomUdpLogger(server='127.0.0.1')

    Arguments
        server (str): UDP server name or IP address.
        port (int): Port number.

    Returns:
        A logger object.

    """

    def __init__(self, server='127.0.0.1', port=514):
        """Create a logging.Logger class with extended functionality."""
        handler = logging.handlers.SysLogHandler((server, port))
        super().__init__(handler=handler)


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    """Customize the Json Formatter."""

    def process_log_record(self, log_record):
        """Add customer record keys and rename threadName key."""
        log_record["version"] = __version__
        log_record["program"] = PROGRAM_NAME
        log_record["service_name"] = log_record.pop('threadName', None)
        # return jsonlogger.JsonFormatter.process_log_record(self, log_record)

        return log_record


class ServiceCheckDiedError(Exception):
    """Raised when a thread that runs service check dies.

    Arguments:
        name (str): the name of the thread
        trace (str): A trace, preferably the output of traceback.format_exc()
    """

    def __init__(self, name, trace):
        """Initialize."""
        super().__init__()
        self.name = name
        self.trace = trace

    def __str__(self):
        """More useful."""
        return ("thread for service {n} died due to : {t}"
                .format(n=self.name, t=self.trace))


def run_custom_bird_reconfigure(operation):
    """Reconfigure BIRD daemon by running a custom command.

    It adds one argument to the command, either "up" or "down".
    If command times out then we kill it. In order to avoid leaving any orphan
    processes, that may have been started by the command, we start a new
    session when we invoke the command and then we kill process group of that
    session.

    Arguments:
        operation (obj): Either a AddOperation or DeleteOperation object.

    """
    log = logging.getLogger(PROGRAM_NAME)
    if isinstance(operation, AddOperation):
        status = 'up'
    else:
        status = 'down'
    cmd = shlex.split(operation.bird_reconfigure_cmd + " " + status)
    log.info("reconfiguring BIRD by running custom command %s", ' '.join(cmd))
    try:
        proc = subprocess.Popen(cmd,
                                start_new_session=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        _, errs = proc.communicate(
            timeout=operation.bird_reconfigure_timeout
        )
    except OSError as exc:
        log.error("reconfiguring BIRD failed with: %s", exc)
    except subprocess.TimeoutExpired as exc:
        log.error("reconfiguring bird timed out")
        if proc.poll() is None:  # if process is still alive
            try:
                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
            except PermissionError as exc:
                log.error("failed to terminate custom bird command: %s", exc)
    else:
        if proc.returncode != 0:
            log.error("reconfiguring BIRD failed with return code: %s and "
                      "stderr: %s", proc.returncode, errs)
        else:
            log.info("custom command successfully reconfigured Bird")



#!/usr/bin/env python3
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
# vim:fenc=utf-8
"""Check if anycast-healthchecker and all configured threads are running.

Usage:
    check_anycast_healthchecker.py [-v]

Options:
    -v   report what it does
"""
import os
import sys
import configparser
import glob
import subprocess
from docopt import docopt


def get_processid(config):
    """Return process id of anycast-healthchecker.

    Arguments:
        config (obj): A configparser object with the configuration of
        anycast-healthchecker.

    Returns:
        The process id found in the pid file

    Raises:
        ValueError in the following cases
        - pidfile option is missing from the configuration
        - pid is either -1 or 1
        - stale pidfile, either with no data or invalid data
        - failure to read pidfile

    """
    pidfile = config.get('daemon', 'pidfile', fallback=None)
    if pidfile is None:
        raise ValueError("Configuration doesn't have pidfile option!")

    try:
        with open(pidfile, 'r') as _file:
            pid = _file.read().rstrip()
            try:
                pid = int(pid)
            except ValueError:
                raise ValueError("stale pid file with invalid data:{}"
                                 .format(pid))
            else:
                if pid in [-1, 1]:
                    raise ValueError("invalid PID ({})".format(pid))
                else:
                    return pid
    except OSError as exc:
        if exc.errno == 2:
            print("CRITICAL: anycast-healthchecker could be down as pid file "
                  "{} doesn't exist".format(pidfile))
            sys.exit(2)
        else:
            raise ValueError("error while reading pid file:{}".format(exc))


def running(pid):
    """Check the validity of a process ID.

    Note: We need root privileges for this to work if we don't run under the same user
    as **anycast-healthechecker**.

    Arguments:
        pid (int): Process ID number.

    Returns:
        True if process ID is found otherwise False.

    """
    try:
        # From kill(2)
        # If sig is 0 (the null signal), error checking is performed but no
        # signal is actually sent. The null signal can be used to check the
        # validity of pid
        os.kill(pid, 0)
    except OSError:
        return False

    return True


def parse_services(config, services):
    """Parse configuration to return number of enabled service checks.

    Arguments:
        config (obj): A configparser object with the configuration of
        anycast-healthchecker.
        services (list): A list of section names which holds configuration
        for each service check

    Returns:
        A number (int) of enabled service checks.

    """
    enabled = 0
    for service in services:
        check_disabled = config.getboolean(service, 'check_disabled')
        if not check_disabled:
            enabled += 1

    return enabled


def main():
    """Run check.

    anycast-healthchecker is a multi-threaded software and for each
    service check it holds a thread. If a thread dies then the service
    is not monitored anymore and the route for the IP associated with service
    it wont be withdrawn in case service goes down in the meantime.
    """
    arguments = docopt(__doc__)
    config_file = '/etc/anycast-healthchecker.conf'
    config_dir = '/etc/anycast-healthchecker.d'
    config = configparser.ConfigParser()
    config_files = [config_file]
    config_files.extend(glob.glob(os.path.join(config_dir, '*.conf')))
    config.read(config_files)

    try:
        pid = get_processid(config)
    except ValueError as exc:
        print("UNKNOWN: {e}".format(e=exc))
        sys.exit(3)
    else:
        process_up = running(pid)

    if not process_up:
        print("CRITICAL: anycast-healthchecker with pid ({p}) isn't running"
              .format(p=pid))
        sys.exit(3)

    services = config.sections()
    services.remove('daemon')
    if not services:
        print("UNKNOWN: No service checks are configured")
        sys.exit(3)

    enabled_service_checks = parse_services(config, services)
    if enabled_service_checks == 0:
        print("OK: Number of service checks is zero, no threads are running")
        sys.exit(0)
    else:
        # parent process plus nummber of threads for each service check
        configured_threads = enabled_service_checks + 1

    cmd = ['/bin/ps', 'h', '-T', '-p', '{n}'.format(n=pid)]
    try:
        if arguments['-v']:
            print("running {}".format(' '.join(cmd)))
        out = subprocess.check_output(cmd, timeout=1)
    except subprocess.CalledProcessError as exc:
        print("UNKNOWN: running '{c}' failed with return code: {r}"
              .format(c=' '.join(cmd), r=exc.returncode))
        sys.exit(3)
    except subprocess.TimeoutExpired:
        print("UNKNOWN: running '{}' timed out".format(' '.join(cmd)))
        sys.exit(3)
    else:
        output_lines = out.splitlines()
        if arguments['-v']:
            for line in output_lines:
                print(line)
        running_threads = len(output_lines)
        if running_threads == configured_threads:
            print("OK: UP (pid={p}) and all threads ({t}) are running"
                  .format(p=pid, t=configured_threads - 1))
            sys.exit(0)
        elif running_threads - 1 == 0:  # minus parent process
            print("CRITICAL: No threads are running OpDocs ANYCAST-03")
            sys.exit(2)
        else:
            print("CRITICAL: Found {n} running threads while configured "
                  "number of threads is {c} OpDocs ANYCAST-03"
                  .format(n=running_threads - 1, c=configured_threads - 1))
            sys.exit(2)


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()



#! /usr/bin/env python
# vim:fenc=utf-8
#
"""A simple healthchecker for Anycasted services.

Usage:
    anycast-healthchecker [ -f <file> -c -p -P ] [ -d <directory> | -F <file> ]

Options:
    -f, --file=<file>          read settings from <file>
                               [default: /etc/anycast-healthchecker.conf]
    -d, --dir=<dir>            read settings for service checks from files
                               under <dir> directory
                               [default: /etc/anycast-healthchecker.d]
    -F, --service-file=<file>  read <file> for settings of a single service
                               check
    -c, --check                perform a sanity check on configuration
    -p, --print                show default settings for anycast-healthchecker
                               and service checks
    -P, --print-conf           show running configuration with default settings
                               applied
    -v, --version              show version
    -h, --help                 show this screen
"""
from functools import partial
import socket
import signal
import sys
from docopt import docopt

from anycast_healthchecker import (DEFAULT_OPTIONS, PROGRAM_NAME, __version__,
                                   healthchecker)
from anycast_healthchecker.utils import (load_configuration, shutdown,
                                         ip_prefixes_sanity_check,
                                         update_pidfile, setup_logger)


def main():
    """Parse CLI and starts main program."""
    args = docopt(__doc__, version=__version__)
    if args['--print']:
        for section in DEFAULT_OPTIONS:
            print("[{}]".format(section))
            for key, value in DEFAULT_OPTIONS[section].items():
                print("{k} = {v}".format(k=key, v=value))
            print()
        sys.exit(0)

    try:
        config, bird_configuration = load_configuration(args['--file'],
                                                        args['--dir'],
                                                        args['--service-file'])
    except ValueError as exc:
        sys.exit('Invalid configuration: ' + str(exc))

    if args['--check']:
        print("OK")
        sys.exit(0)

    if args['--print-conf']:
        for section in config:
            print("[{}]".format(section))
            for key, value in config[section].items():
                print("{k} = {v}".format(k=key, v=value))
            print()
        sys.exit(0)

    try:
        lock_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        lock_socket.bind('\0' + "{}".format(PROGRAM_NAME))
    except socket.error as exc:
        sys.exit("failed to acquire a lock by creating an abstract namespace"
                 " socket: {}".format(exc))
    else:
        print("acquired a lock by creating an abstract namespace socket: {}"
              .format(lock_socket))

    # Clean old pidfile, if it exists, and write PID to it.
    pidfile = config.get('daemon', 'pidfile')
    update_pidfile(pidfile)

    # Register our shutdown handler to various termination signals.
    shutdown_handler = partial(shutdown, pidfile)
    signal.signal(signal.SIGHUP, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)
    signal.signal(signal.SIGABRT, shutdown_handler)
    signal.signal(signal.SIGINT, shutdown_handler)

    # Set up loggers.
    logger = setup_logger(config)

    # Perform a sanity check on IP-Prefixes
    ip_prefixes_sanity_check(config, bird_configuration)

    # Create our master process.
    checker = healthchecker.HealthChecker(config, bird_configuration)
    logger.info("starting %s version %s", PROGRAM_NAME, __version__)
    checker.run()


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()
